""" Test for the evaluation of the ahb expression conditions tests (Mussfeldprüfung) """
import uuid
from typing import List
from unittest.mock import AsyncMock

import inject
import pytest  # type:ignore[import]
import pytest_asyncio  # type:ignore[import]

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableDataProvider
from ahbicht.content_evaluation.evaluator_factory import create_and_inject_hardcoded_evaluators
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider
from ahbicht.evaluation_results import (
    AhbExpressionEvaluationResult,
    FormatConstraintEvaluationResult,
    RequirementConstraintEvaluationResult,
)
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint
from ahbicht.expressions.enums import ModalMark, PrefixOperator, RequirementIndicator
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from unittests.defaults import (
    default_test_format,
    default_test_version,
    empty_default_hints_provider,
    empty_default_rc_evaluator,
    iterating_rc_evaluator,
    return_empty_dummy_evaluatable_data,
)


class TestAHBExpressionEvaluation:
    """Test for the evaluation of the ahb expression conditions tests (Mussfeldprüfung)"""

    @pytest_asyncio.fixture()
    def setup_and_teardown_injector(self):
        inject.clear_and_configure(
            lambda binder: binder.bind(
                TokenLogicProvider,
                SingletonTokenLogicProvider([empty_default_hints_provider, empty_default_rc_evaluator]),
            ).bind_to_provider(EvaluatableDataProvider, return_empty_dummy_evaluatable_data)
        )
        yield
        inject.clear()

    @pytest_asyncio.fixture()
    def setup_and_teardown_injector_without_evaluatable_data_provider(self):
        inject.clear_and_configure(
            lambda binder: binder.bind(
                TokenLogicProvider,
                SingletonTokenLogicProvider([empty_default_hints_provider, empty_default_rc_evaluator]),
            )
            # similar to the fixture above but without the evaluatable data provider => leads to injection errors
        )
        yield
        inject.clear()

    @pytest_asyncio.fixture()
    def setup_and_teardown_injector_with_iterating_rc_evaluator(self):
        inject.clear_and_configure(
            lambda binder: binder.bind(
                TokenLogicProvider,
                SingletonTokenLogicProvider([empty_default_hints_provider, iterating_rc_evaluator]),
            ).bind_to_provider(EvaluatableDataProvider, return_empty_dummy_evaluatable_data)
        )
        yield
        inject.clear()

    async def test_evaluation_under_cache(self, setup_and_teardown_injector_with_iterating_rc_evaluator):
        """
        This test is to show that the content evaluation is _not_ affected by the cache, meaning it's still possible
        to get different outcomes for the same expression even if it has been evaluated once already.
        This is a very basic assertion but not always given.
        """
        expression = "Muss [1] U [2]"
        evaluation_results: List[AhbExpressionEvaluationResult] = []
        for _ in range(9):
            # see iterating_rc_evaluator for the behaviour of 1 and 2. Both iterate
            # 1: (true, false, true, false, ...)
            # 2: (true, true, false, false, true, true, false, false, ...)
            # --------------------------
            #   1   |   2   | [1] U [2]
            # true  | true  | true
            # false | true  | false
            # true  | false | false
            # false | false | false
            # true  | true  | true
            # ... repeat
            tree = parse_ahb_expression_to_single_requirement_indicator_expressions(expression)
            evaluation_result = await evaluate_ahb_expression_tree(tree)
            evaluation_results.append(evaluation_result)
        overall_results = [
            x.requirement_constraint_evaluation_result.requirement_constraints_fulfilled for x in evaluation_results
        ]
        assert overall_results == [True, False, False, False, True, False, False, False, True]

    @pytest.mark.parametrize(
        """ahb_expression, expected_requirement_indicator, expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional, expected_format_constraints_expression, expected_hints""",
        [
            pytest.param("Muss", ModalMark.MUSS, True, False, None, None),
            pytest.param("X", PrefixOperator.X, True, False, None, None),
            pytest.param("Muss[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("muss[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("M[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("m[1]", ModalMark.MUSS, True, True, None, None),
            pytest.param("Soll[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("soll[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("S[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("s[1]", ModalMark.SOLL, True, True, None, None),
            pytest.param("Kann[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("kann[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("K[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("k[1]", ModalMark.KANN, True, True, None, None),
            pytest.param("Muss[2]", ModalMark.MUSS, False, True, None, None),
            pytest.param("Muss[1]Soll[2]", ModalMark.MUSS, True, True, None, None),
            pytest.param("Muss[2]Soll[1]Kann[4]", ModalMark.SOLL, True, True, None, None),
            pytest.param("Muss[2]\tSoll [ 1]  Kann[4\t]", ModalMark.SOLL, True, True, None, None),
            pytest.param("X[1]", PrefixOperator.X, True, True, None, None),
            pytest.param("O[2]", PrefixOperator.O, False, True, None, None),
            pytest.param("U[1]", PrefixOperator.U, True, True, None, None),
            pytest.param("U[1]U[2]", PrefixOperator.U, False, True, None, None),
            pytest.param("X[1]U[2]U[3]O[4]", PrefixOperator.X, False, True, None, None),
            pytest.param("Muss([1]O[2])U[3]Soll[2]Kann[2]O[4]", ModalMark.MUSS, True, True, None, None),
            pytest.param("muss([1]o[2])u[3]soll[2]kann[2]O[4]", ModalMark.MUSS, True, True, None, None),
            pytest.param("Muss[2]Soll[2]", ModalMark.SOLL, False, True, None, None),
            # Neutral value
            pytest.param("Muss[503]", ModalMark.MUSS, True, False, None, "[503]"),
            pytest.param("Soll[902]", ModalMark.SOLL, True, False, "[902]", None),
            pytest.param("Kann[503]U[504]", ModalMark.KANN, True, False, None, "[503]U[504]"),
            # Last modal mark without conditions
            pytest.param("Muss[1]Kann", ModalMark.MUSS, True, True, None, None),
            pytest.param("Muss[2]Kann", ModalMark.KANN, True, True, None, None),
        ],
    )
    async def test_evaluate_valid_ahb_expression(
        self,
        mocker,
        ahb_expression,
        expected_requirement_indicator: RequirementIndicator,
        expected_requirement_constraints_fulfilled,
        expected_requirement_is_conditional,
        expected_format_constraints_expression,
        expected_hints,
    ):
        """
        Tests that valid ahb expressions are evaluated as expected.
        Odd condition_keys are True, even condition_keys are False
        """

        def side_effect_rc_evaluation(condition_expression):
            if condition_expression.lower() in ["[1]", " [ 1]  ", "[3]", "([1]o[2])u[3]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=True,
                    format_constraints_expression=None,
                    hints=None,
                )
            if condition_expression.lower() in [
                "[2]",
                "[2]\t",
                "[4]",
                "[4\t]",
                "[1]u[2]",
                "[1]u[2]u[3]o[4]",
                "[2]o[4]",
            ]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=False,
                    requirement_is_conditional=True,
                    format_constraints_expression=None,
                    hints=None,
                )
            if condition_expression.lower() in ["[503]", "[503]u[504]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=False,
                    format_constraints_expression=None,
                    hints=condition_expression,
                )
            if condition_expression.lower() in ["[902]"]:
                return RequirementConstraintEvaluationResult(
                    requirement_constraints_fulfilled=True,
                    requirement_is_conditional=False,
                    format_constraints_expression=condition_expression,
                    hints=None,
                )

        mocker.patch(
            "ahbicht.expressions.ahb_expression_evaluation.requirement_constraint_evaluation",
            side_effect=AsyncMock(side_effect=side_effect_rc_evaluation),
        )
        mocker.patch(
            "ahbicht.expressions.ahb_expression_evaluation.format_constraint_evaluation",
            return_value=AsyncMock(
                side_effect=FormatConstraintEvaluationResult(format_constraints_fulfilled=True, error_message=None)
            ),
        )

        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        result = await evaluate_ahb_expression_tree(parsed_tree)

        assert result.requirement_indicator == expected_requirement_indicator
        assert (
            result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            == expected_requirement_constraints_fulfilled
        )
        assert (
            result.requirement_constraint_evaluation_result.requirement_is_conditional
            == expected_requirement_is_conditional
        )
        assert (
            result.requirement_constraint_evaluation_result.format_constraints_expression
            == expected_format_constraints_expression
        )
        assert result.requirement_constraint_evaluation_result.hints == expected_hints

    @pytest.mark.parametrize(
        "expression, expected_error, expected_error_message",
        [
            pytest.param(
                "Muss[]",
                SyntaxError,
                """Please make sure that:
             * all conditions have the form [INT]
             * all packages have the form [INTPn..m]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """,
            ),
            pytest.param(
                "Soll[1001]",
                ValueError,
                """Condition key is not in valid number range.""",
            ),
        ],
    )
    async def test_not_wellformed_ahb_expressions(
        self, expression: str, expected_error: type, expected_error_message: str, setup_and_teardown_injector
    ):
        """Tests that an error is raised when trying to pass invalid values."""

        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(expression)

        with pytest.raises(expected_error) as excinfo:  # type: ignore[var-annotated]
            await evaluate_ahb_expression_tree(parsed_tree)

        assert expected_error_message in str(excinfo.value)

    async def test_missing_bind_to_provider_error(self, setup_and_teardown_injector_without_evaluatable_data_provider):
        """Tests that a meaningful error raised when the user forgot to setup bind_to_provider"""

        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions("Muss [1] U [2]")
        with pytest.raises(AttributeError) as excinfo:  # type: ignore[var-annotated]
            await evaluate_ahb_expression_tree(parsed_tree)
        assert "Are you sure you called .bind_to_provider" in str(excinfo.value)

    @pytest.mark.parametrize(
        "ahb_expression, content_evaluation_result",
        [
            pytest.param(
                "Muss ([2]O[3])[902][501]",
                ContentEvaluationResult(
                    hints={"501": "foo"},
                    format_constraints={
                        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                    },
                    requirement_constraints={
                        "2": ConditionFulfilledValue.FULFILLED,
                        "3": ConditionFulfilledValue.UNFULFILLED,
                    },
                    id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
                    packages={},
                ),
            )
        ],
    )
    async def test_all_serializations_work_similar(
        self, ahb_expression: str, content_evaluation_result: ContentEvaluationResult
    ):
        tree_a = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        tree_b = await parse_expression_including_unresolved_subexpressions(ahb_expression)
        # it's OK/expected that the trees look different depending on whether sub expressions are resolved or not
        # but in any case the evaluation result should look the same
        try:
            create_and_inject_hardcoded_evaluators(
                content_evaluation_result,
                evaluatable_data_provider=return_empty_dummy_evaluatable_data,
                edifact_format=default_test_format,
                edifact_format_version=default_test_version,
            )
            evaluation_result_a = await evaluate_ahb_expression_tree(tree_a)
            evaluation_result_b = await evaluate_ahb_expression_tree(tree_b)
            assert evaluation_result_a == evaluation_result_b
        finally:
            inject.clear()

    @pytest.mark.parametrize(
        "ahb_expression, content_evaluation_result",
        [
            pytest.param(
                "Muss [901][1] X [902][2]",
                ContentEvaluationResult(
                    hints={"501": "foo"},
                    format_constraints={
                        "901": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                    },
                    requirement_constraints={
                        "1": ConditionFulfilledValue.FULFILLED,
                        "2": ConditionFulfilledValue.UNFULFILLED,
                    },
                    id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
                    packages={},
                ),
            ),
            pytest.param(
                "Muss [901][1] X [902][2]",
                ContentEvaluationResult(
                    hints={"501": "foo"},
                    format_constraints={
                        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                        "901": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                    },
                    requirement_constraints={
                        "2": ConditionFulfilledValue.FULFILLED,
                        "1": ConditionFulfilledValue.UNFULFILLED,
                    },
                    id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
                    packages={},
                ),
            ),
        ],
    )
    async def test_valid_expression_evaluation(
        self, ahb_expression: str, content_evaluation_result: ContentEvaluationResult
    ):
        tree = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        try:
            create_and_inject_hardcoded_evaluators(
                content_evaluation_result,
                evaluatable_data_provider=return_empty_dummy_evaluatable_data,
                edifact_format=default_test_format,
                edifact_format_version=default_test_version,
            )
            evaluation_result = await evaluate_ahb_expression_tree(tree)
            assert evaluation_result is not None
        finally:
            inject.clear()

    @pytest.mark.parametrize(
        "ahb_expression, content_evaluation_result, expected",
        [
            pytest.param(
                "Soll [1]",
                ContentEvaluationResult(
                    hints={},
                    format_constraints={},
                    requirement_constraints={
                        "1": ConditionFulfilledValue.UNKNOWN,
                    },
                    id=uuid.UUID("9bdc494b-8e61-440b-a15d-eb5630916969"),
                    packages={},
                ),
                AhbExpressionEvaluationResult(
                    requirement_indicator=ModalMark.SOLL,
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        requirement_constraints_fulfilled=None,
                        requirement_is_conditional=None,
                        format_constraints_expression=None,
                        hints=None,
                    ),
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        format_constraints_fulfilled=True
                    ),
                ),
            ),
            pytest.param(
                "Muss [1]",
                ContentEvaluationResult(
                    hints={},
                    format_constraints={},
                    requirement_constraints={
                        "1": ConditionFulfilledValue.UNKNOWN,
                    },
                    id=uuid.UUID("ba79e51c-6b74-44ed-ad53-66fba828b1b8"),
                    packages={},
                ),
                AhbExpressionEvaluationResult(
                    requirement_indicator=ModalMark.MUSS,
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        requirement_constraints_fulfilled=None,
                        requirement_is_conditional=None,
                        format_constraints_expression=None,
                        hints=None,
                    ),
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        format_constraints_fulfilled=True
                    ),
                ),
            ),
        ],
    )
    async def test_no_not_implemented_error_is_raised_for_unknown_nodes(
        self,
        ahb_expression: str,
        content_evaluation_result: ContentEvaluationResult,
        expected: AhbExpressionEvaluationResult,
    ):
        tree = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)
        try:
            create_and_inject_hardcoded_evaluators(
                content_evaluation_result,
                evaluatable_data_provider=return_empty_dummy_evaluatable_data,
                edifact_format=default_test_format,
                edifact_format_version=default_test_version,
            )
            evaluation_result = await evaluate_ahb_expression_tree(tree)
            assert evaluation_result is not None
            assert evaluation_result == expected
        finally:
            inject.clear()



"""
This module contains classes that are returned by mappers, meaning they contain a mapping.
"""
import re
from typing import Match, Optional

import attrs
from efoli import EdifactFormat
from marshmallow import Schema, fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True)
class ConditionKeyConditionTextMapping:
    """
    maps a condition from a specified EDIFACT format onto a text as it is found in the AHB.
    """

    edifact_format: EdifactFormat = attrs.field(
        validator=attrs.validators.instance_of(EdifactFormat)
    )  #: the format in which the condition is used; e.g. 'UTILMD'
    condition_key: str = attrs.field(
        validator=attrs.validators.instance_of(str)
    )  #: the key of the condition without square brackets; e.g. '78'
    condition_text: Optional[str] = attrs.field(default=None)
    """
    the description of the condition as in the AHB; None if unknown;
    e.g. 'Wenn SG4 STS+7++E02 (Transaktionsgrund: Einzug/Neuanlage)  nicht vorhanden'.
    """


class ConditionKeyConditionTextMappingSchema(Schema):
    """
    A schema to (de-)serialize :class:`.ConditionKeyConditionTextMapping` s
    """

    edifact_format = EnumField(EdifactFormat)
    condition_key = fields.String()
    condition_text = fields.String(load_default=None)

    # pylint:disable=unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> ConditionKeyConditionTextMapping:
        """
        Converts the barely typed data dictionary into an actual :class:`.ConditionKeyConditionTextMapping`
        """
        return ConditionKeyConditionTextMapping(**data)


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True)
class PackageKeyConditionExpressionMapping:
    """
    maps a package key from a specified EDIFACT format onto a (not yet parsed) condition expression as it is found in
    the AHB.
    """

    edifact_format: EdifactFormat = attrs.field(
        validator=attrs.validators.instance_of(EdifactFormat)
    )  #: the format in which the package is used; e.g. 'UTILMD'
    package_key: str = attrs.field(
        validator=attrs.validators.instance_of(str)
    )  #: the key of the package without square brackets but with trailing P; e.g. '10P'
    package_expression: Optional[str] = attrs.field(
        default=None
    )  #: the expression for which the package is a shortcut; None if unknown e.g. '[20] ∧ [244]'

    def has_been_resolved_successfully(self) -> bool:
        """
        return true iff the package has been resolved successfully
        """
        return self.package_expression is not None


class PackageKeyConditionExpressionMappingSchema(Schema):
    """
    A schema to (de-)serialize :class:`.PackageKeyConditionExpressionMapping` s
    """

    edifact_format = EnumField(EdifactFormat)
    package_key = fields.String()
    package_expression = fields.String(load_default=None)

    # pylint:disable=unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> PackageKeyConditionExpressionMapping:
        """
        Converts the barely typed data dictionary into an actual :class:`.PackageKeyConditionExpressionMapping`
        """
        return PackageKeyConditionExpressionMapping(**data)


# pylint:disable=unused-argument
def check_max_greater_or_equal_than_min(instance: "Repeatability", attribute, value):
    """
    assert that 0<=min<max and not both min and max are 0
    """
    if not 0 <= instance.min_occurrences <= instance.max_occurrences:
        raise ValueError(f"0≤n≤m is not fulfilled for n={instance.min_occurrences}, m={instance.max_occurrences}")
    if instance.min_occurrences == instance.max_occurrences == 0:
        raise ValueError("not both min and max occurrences must be 0")


# pylint:disable=too-few-public-methods
@attrs.define(auto_attribs=True, kw_only=True)
class Repeatability:
    """
    describes how often a segment/code must be used when a "repeatability" is provided with packages
    """

    min_occurrences: int = attrs.field(
        validator=attrs.validators.and_(attrs.validators.instance_of(int), check_max_greater_or_equal_than_min)
    )
    """
    how often the segment/code has to be repeated (lower, inclusive bound); may be 0 for optional packages
    """

    max_occurrences: int = attrs.field(
        validator=attrs.validators.and_(attrs.validators.instance_of(int), check_max_greater_or_equal_than_min)
    )
    """
    how often the segment/coode may be repeated at most (upper, inclusive bound).
    This is inclusive meaning that [123P0..1] leads to max_occurrences==1
    """

    def is_optional(self) -> bool:
        """
        returns true if the package used together with this repeatability is optional
        """
        return self.min_occurrences == 0


_repeatability_pattern = re.compile(r"^(?P<min>\d+)\.{2}(?P<max>\d+)$")  #: a pattern to match "n..m" repeatabilities


def parse_repeatability(repeatability_string: str) -> Repeatability:
    """
    parses the given string as repeatability; e.g. `17..23` is parsed as min=17, max=23
    """
    match: Optional[Match[str]] = _repeatability_pattern.match(repeatability_string)
    if match is None:
        raise ValueError(f"The given string '{repeatability_string}' could not be parsed as repeatability")
    min_repeatability = int(match["min"])
    max_repeatability = int(match["max"])
    return Repeatability(min_occurrences=min_repeatability, max_occurrences=max_repeatability)



import inject
import pytest  # type:ignore[import]
import pytest_asyncio  # type:ignore[import]

from unittests.defaults import EmptyDefaultRcEvaluator


class SomethingInjectable:
    """
    This class is used by the custom Evaluators below
    """


class RcEvaluatorThatUsesInjectAttr(EmptyDefaultRcEvaluator):
    # this won't work and the test proves it
    something_injectable: SomethingInjectable = inject.attr(SomethingInjectable)


class TestErrorWhenUsingInjectAttr:
    """
    Asserts on a specific error message when a library user tries to use inject.attr
    """

    @pytest_asyncio.fixture()
    def clear_injectors(self):
        # This might not be necessary when you only run this test alone but without this fixture
        # running all the tests (e.g. via tox) might fail
        inject.clear()

    def test_attribute_error_with_inject_attr(self, clear_injectors):
        """
        This tests that an error with a meaningful error message is raised.
        Test can be removed as soon as the underlying bug in inject is resolved.
        """
        with pytest.raises(AttributeError) as attribute_error:
            RcEvaluatorThatUsesInjectAttr()
        assert (
            "Using inject.attr in custom evaluators as you tried in RcEvaluatorThatUsesInjectAttr is not supported."
            in str(attribute_error)
        )



"""
Package Expansion is the process of finding the condition expression which was abbreviated by using a package.
e.g. if inside a tree "[123P]" is replaced by "[1] U ([2] O [3])".
"""
import json
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Mapping, Optional

import inject
from efoli import EdifactFormat, EdifactFormatVersion

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluatableDataProvider
from ahbicht.mapping_results import PackageKeyConditionExpressionMapping, PackageKeyConditionExpressionMappingSchema


# pylint:disable=too-few-public-methods
class PackageResolver(ABC):
    """
    A package resolver provides condition expressions for given package keys.
    """

    # define some common attributes. They will be needed to find the correct resolver for each use case.
    edifact_format: EdifactFormat = NotImplementedError(  # type:ignore[assignment]
        "The inheriting package resolver needs to define a format to which it is applicable."
    )  #: the format for which the resolver may be used
    edifact_format_version: EdifactFormatVersion = NotImplementedError(  # type:ignore[assignment]
        "The inheriting package resolver needs to define a format version."
    )  #: the format version for which the resolver may be used

    def __init__(self):
        self.logger = logging.getLogger(self.__module__)
        self.logger.setLevel(logging.DEBUG)
        self.logger.info("Instantiated %s", self.__class__.__name__)

    @abstractmethod
    async def get_condition_expression(self, package_key: str) -> PackageKeyConditionExpressionMapping:
        """
        Returns a condition expression (e.g. "[1] U ([2] O [3])") for the given package_key (e.g. "123P")
        Returns None in the package_expression if the package is unresolvable (see 'has_been_resolved_successfully').
        :param package_key: The unique (integer) key of the package. The 'P' suffix is required.
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method.")


class DictBasedPackageResolver(PackageResolver):
    """
    A Package Resolver that is based on hardcoded values from a dictionary
    """

    def __init__(self, results: Mapping[str, Optional[str]]):
        """
        Initialize with a dictionary that contains all the condition expressions.
        :param results: maps the package key (e.g. '123') to the package expression (e.g. '[1] U [2]')
        """
        super().__init__()
        for key in results.keys():
            if not key.endswith("P"):
                raise ValueError("The keys should end with 'P' to avoid ambiguities. Use '123P' instead of '123'.")
        self._all_packages: Mapping[str, Optional[str]] = results

    async def get_condition_expression(self, package_key: str) -> PackageKeyConditionExpressionMapping:
        if not package_key:
            raise ValueError(f"The package key must not be None/empty but was '{package_key}'")
        if not package_key.endswith("P"):
            raise ValueError("The package key should be provided with a trailing 'P'.")
        result: PackageKeyConditionExpressionMapping
        if package_key in self._all_packages:
            result = PackageKeyConditionExpressionMapping(
                package_key=package_key,
                package_expression=self._all_packages[package_key],
                edifact_format=EdifactFormat.UTILMD,
            )
        else:
            result = PackageKeyConditionExpressionMapping(
                package_key=package_key,
                package_expression=None,
                edifact_format=EdifactFormat.UTILMD,
            )
        self.logger.debug("Resolved expression '%s' for package key %s", result.package_expression, result.package_key)
        return result


class JsonFilePackageResolver(DictBasedPackageResolver):
    """
    The JsonFilePackageResolver loads package keys/expressions from a JSON file.
    """

    def __init__(self, edifact_format: EdifactFormat, edifact_format_version: EdifactFormatVersion, file_path: Path):
        super().__init__(self._open_and_load_package_mappings(file_path))
        self.edifact_format = edifact_format
        self.edifact_format_version = edifact_format_version

    @staticmethod
    def _open_and_load_package_mappings(file_path: Path) -> Dict[str, Optional[str]]:
        """
        Opens the hint json file and loads it into an attribute of the class.
        The method can read both a dictionary of package key/package expression mappings and a
        list of PackageKeyConditionExpressionMappings
        """
        with open(file_path, encoding="utf-8") as json_infile:
            json_body = json.load(json_infile)
        if isinstance(json_body, dict):
            # {"1P": "[2] U [3]", "2P": "[4] O [5]"...
            return json_body
        # [{PackageKeyConditionExpressionMapping},...]
        mapping_list: List[PackageKeyConditionExpressionMapping] = PackageKeyConditionExpressionMappingSchema().load(
            json_body, many=True
        )
        return {mapping.package_key: mapping.package_expression for mapping in mapping_list}


class ContentEvaluationResultBasedPackageResolver(PackageResolver):
    """
    A package resolver that expects the evaluatable data to contain a ContentEvalutionResult as edifact seed.
    Other than the DictBasedPackageResolver the outcome is not dependent on the initialization but on the
    evaluatable data.
    """

    def __init__(self):
        super().__init__()
        self._schema = ContentEvaluationResultSchema()

    async def get_condition_expression(self, package_key: str) -> PackageKeyConditionExpressionMapping:
        # the missing second argument to the private method call in the next line should be injected automatically
        return await self._get_condition_expression(package_key)  # pylint:disable=no-value-for-parameter

    @inject.params(evaluatable_data=EvaluatableDataProvider)  # injects what has been bound to the EvaluatableData type
    async def _get_condition_expression(
        self, package_key: str, evaluatable_data: EvaluatableData
    ) -> PackageKeyConditionExpressionMapping:
        content_evaluation_result: ContentEvaluationResult = self._schema.load(evaluatable_data.body)
        try:
            self.logger.debug("Retrieving package '%s' from Content Evaluation Result", package_key)
            if content_evaluation_result.packages is None:
                content_evaluation_result.packages = {}
            package_expression = content_evaluation_result.packages[package_key]
            return PackageKeyConditionExpressionMapping(
                edifact_format=self.edifact_format, package_expression=package_expression, package_key=package_key
            )
        except KeyError as key_error:
            self.logger.debug("Package '%s' was not contained in the CER", str(key_error))
            return PackageKeyConditionExpressionMapping(
                edifact_format=self.edifact_format,
                package_expression=None,
                package_key=package_key,
            )



from typing import List

import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]
from efoli import EdifactFormat, EdifactFormatVersion

from ahbicht.content_evaluation.evaluators import Evaluator
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider, TokenLogicProvider


class TestEvaluatorProvider:
    """
    Test that the list based evalutor provider works as expected
    """

    def test_initialization(self):
        evaluators: List[Evaluator] = []
        # setup some test data/instances
        for edifact_format in EdifactFormat:
            if edifact_format == EdifactFormat.COMDIS:
                # we use the comdis format to test the "not found"/not implemented case below
                continue
            for edifact_format_version in EdifactFormatVersion:

                class ExampleRcEvaluator(RcEvaluator):
                    def _get_default_context(self):
                        return None

                    def __init__(self):
                        super().__init__()
                        self.edifact_format_version = edifact_format_version
                        self.edifact_format = edifact_format

                class ExampleFcEvaluator(FcEvaluator):
                    def __init__(self):
                        super().__init__()
                        self.edifact_format_version = edifact_format_version
                        self.edifact_format = edifact_format

                example_rc_evaluator = ExampleRcEvaluator()
                example_fc_evaluator = ExampleFcEvaluator()
                evaluators.append(example_rc_evaluator)
                evaluators.append(example_fc_evaluator)
        # here's where the setup is over and the actual test begins
        evaluator_provider: TokenLogicProvider = SingletonTokenLogicProvider(evaluators)
        for edifact_format in EdifactFormat:
            for edifact_format_version in EdifactFormatVersion:
                if edifact_format == EdifactFormat.COMDIS:
                    # comdis is not set up (see above)
                    with pytest.raises(NotImplementedError):
                        _ = evaluator_provider.get_fc_evaluator(edifact_format, edifact_format_version)
                    with pytest.raises(NotImplementedError):
                        _ = evaluator_provider.get_rc_evaluator(edifact_format, edifact_format_version)
                    continue
                fc_evaluator = evaluator_provider.get_fc_evaluator(edifact_format, edifact_format_version)
                rc_evaluator = evaluator_provider.get_rc_evaluator(edifact_format, edifact_format_version)
                assert isinstance(fc_evaluator, FcEvaluator)
                assert isinstance(rc_evaluator, RcEvaluator)
                assert fc_evaluator.edifact_format_version == edifact_format_version
                assert fc_evaluator.edifact_format == edifact_format
                assert rc_evaluator.edifact_format_version == edifact_format_version
                assert rc_evaluator.edifact_format == edifact_format



"""
Contains a class that is able to provide any RC/FC Evaluator, HintsProvider or Package Resolver.
"""
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Union

from efoli import EdifactFormat, EdifactFormatVersion

from ahbicht.content_evaluation.evaluators import Evaluator
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.content_evaluation.rc_evaluators import RcEvaluator
from ahbicht.expressions.hints_provider import HintsProvider
from ahbicht.expressions.package_expansion import PackageResolver


class TokenLogicProvider(ABC):
    """
    A TokenLogicProvider is a class that can provide you with the correct evaluator/hints provider/package resolver
    for your use case/for any token in the parsed expression tree.
    """

    @abstractmethod
    def get_rc_evaluator(self, edifact_format: EdifactFormat, format_version: EdifactFormatVersion) -> RcEvaluator:
        """
        returns an appropriate RC Evaluator for the given edifact_format and edifact_format_version.
        The implementing class shall raise a NotImplementedError and not return None.
        """
        raise NotImplementedError("The inheriting sub class has to implement this method")

    @abstractmethod
    def get_fc_evaluator(self, edifact_format: EdifactFormat, format_version: EdifactFormatVersion) -> FcEvaluator:
        """
        returns an appropriate FC Evaluator for the given edifact_format and edifact_format_version.
        The implementing class shall raise a NotImplementedError and not return None.
        """
        raise NotImplementedError("The inheriting sub class has to implement this method")

    @abstractmethod
    def get_hints_provider(
        self, edifact_format: Optional[EdifactFormat] = None, format_version: Optional[EdifactFormatVersion] = None
    ) -> HintsProvider:
        """
        returns an appropriate HintsProvider for the given edifact_format and edifact_format_version.
        The implementing class shall raise a NotImplementedError and not return None.
        """
        raise NotImplementedError("The inheriting sub class has to implement this method")

    @abstractmethod
    def get_package_resolver(
        self, edifact_format: Optional[EdifactFormat] = None, format_version: Optional[EdifactFormatVersion] = None
    ) -> PackageResolver:
        """
        Returns an appropriate PackageResolver for the given edifact_format and edifact_format_version.
        You can provide None as format(version) if there is only 1 package resolver available
        The implementing class shall raise a NotImplementedError and not return None.
        """
        raise NotImplementedError("The inheriting sub class has to implement this method")


class SingletonTokenLogicProvider(TokenLogicProvider):
    """
    An TokenLogicProvider that is instantiated with a list of evaluators/providers/resolvers of which the same instances
    will be used during the entire uptime of the application (singleton style).
    """

    _unknown_key = "undefined"

    @staticmethod
    def _to_key(edifact_format: Optional[EdifactFormat], format_version: Optional[EdifactFormatVersion]) -> str:
        """
        because a tuple for format and format version is not hashable / usuable as key in a dict this methods
        converts them to a unique and hashable string
        """
        if edifact_format is None or format_version is None:
            return SingletonTokenLogicProvider._unknown_key
        # we don't care what the key is, it just has to be unique and consistent
        return f"{edifact_format}-{format_version}"

    def __init__(self, inputs: List[Union[Evaluator, PackageResolver, HintsProvider]]):
        self._rc_evaluators: Dict[str, RcEvaluator] = {}
        self._fc_evaluators: Dict[str, FcEvaluator] = {}
        self._hints_providers: Dict[str, HintsProvider] = {}
        self._package_resolvers: Dict[str, PackageResolver] = {}
        for instance in inputs:
            key: str
            try:
                key = SingletonTokenLogicProvider._to_key(instance.edifact_format, instance.edifact_format_version)
            except NotImplementedError:
                # this is ok, if there's only one of the kind (e.g. only one package resolver, only one hints provider)
                # if the user tries to provide more than 1 instance of the same kind without specifying format(version)
                # they'll run into an value error below
                key = SingletonTokenLogicProvider._unknown_key
            target_dict: Dict[str, Any]
            if isinstance(instance, RcEvaluator):
                target_dict = self._rc_evaluators
            elif isinstance(instance, FcEvaluator):
                target_dict = self._fc_evaluators
            elif isinstance(instance, HintsProvider):
                target_dict = self._hints_providers
            elif isinstance(instance, PackageResolver):
                target_dict = self._package_resolvers
            else:
                raise ValueError(f"The type of '{instance}' is not supported. Expected either RC or FC evaluator")
            if key in target_dict:
                conflict = target_dict[key]
                raise ValueError(
                    f"The key '{key}' is already used by {conflict}. For this reason you cannot add '{instance}'"
                )
            target_dict[key] = instance

    def get_fc_evaluator(
        self, edifact_format: Optional[EdifactFormat] = None, format_version: Optional[EdifactFormatVersion] = None
    ) -> FcEvaluator:
        try:
            return self._fc_evaluators[SingletonTokenLogicProvider._to_key(edifact_format, format_version)]
        except KeyError as key_error:
            raise NotImplementedError(
                f"No FC Evaluator has been registered for {edifact_format} in {format_version}"
            ) from key_error

    def get_rc_evaluator(
        self, edifact_format: Optional[EdifactFormat] = None, format_version: Optional[EdifactFormatVersion] = None
    ) -> RcEvaluator:
        try:
            return self._rc_evaluators[SingletonTokenLogicProvider._to_key(edifact_format, format_version)]
        except KeyError as key_error:
            raise NotImplementedError(
                f"No RC Evaluator has been registered for {edifact_format} in {format_version}"
            ) from key_error

    def get_hints_provider(
        self, edifact_format: Optional[EdifactFormat] = None, format_version: Optional[EdifactFormatVersion] = None
    ) -> HintsProvider:
        try:
            return self._hints_providers[SingletonTokenLogicProvider._to_key(edifact_format, format_version)]
        except KeyError as key_error:
            raise NotImplementedError(
                f"No HintsProvider has been registered for {edifact_format} in {format_version}"
            ) from key_error

    def get_package_resolver(
        self, edifact_format: Optional[EdifactFormat] = None, format_version: Optional[EdifactFormatVersion] = None
    ) -> PackageResolver:
        try:
            return self._package_resolvers[SingletonTokenLogicProvider._to_key(edifact_format, format_version)]
        except KeyError as key_error:
            raise NotImplementedError(
                f"No PackageResolver has been registered for {edifact_format} in {format_version}"
            ) from key_error



"""
why conftest.py?
https://docs.pytest.org/en/7.1.x/how-to/fixtures.html#scope-sharing-fixtures-across-classes-modules-packages-or-session
"""
import inject
import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]

from ahbicht.content_evaluation import ContentEvaluationResult, EvaluatableData, TokenLogicProvider
from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResultSchema
from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableDataProvider
from ahbicht.content_evaluation.evaluator_factory import create_content_evaluation_result_based_evaluators
from ahbicht.content_evaluation.token_logic_provider import SingletonTokenLogicProvider
from unittests.defaults import default_test_format, default_test_version


def store_content_evaluation_result_in_evaluatable_data(
    content_evaluation_result: ContentEvaluationResult,
) -> EvaluatableData[dict]:
    """
    a helper method for the tests to store a serialized content evaluation result in an EvaluatableData instance
    :param content_evaluation_result:
    :return: a new EvaluatableData instance
    """
    schema = ContentEvaluationResultSchema()
    cer_dict = schema.dump(content_evaluation_result)
    return EvaluatableData(
        body=cer_dict, edifact_format=default_test_format, edifact_format_version=default_test_version
    )


@pytest.fixture
def inject_cer_evaluators(request: SubRequest):
    content_evaluation_result: ContentEvaluationResult = request.param
    assert isinstance(content_evaluation_result, ContentEvaluationResult)

    def get_evaluatable_data():
        return store_content_evaluation_result_in_evaluatable_data(content_evaluation_result)

    def configure(binder):
        binder.bind(
            TokenLogicProvider,
            SingletonTokenLogicProvider(
                [*create_content_evaluation_result_based_evaluators(default_test_format, default_test_version)]
            ),
        )
        binder.bind_to_provider(EvaluatableDataProvider, get_evaluatable_data)

    inject.configure_once(configure)
    yield
    inject.clear()



import pytest  # type:ignore[import]
from maus.models.edifact_components import DataElementDataType

from ahbicht.validation.validation_results import (
    DataElementValidationResult,
    ListOfValidationResultInContext,
    SegmentLevelValidationResult,
    ValidationResultInContext,
)
from ahbicht.validation.validation_values import RequirementValidationValue

# lovric = list_of_validation_result_in_context


class TestValidationResults:
    @pytest.mark.parametrize(
        "lovric_actual, edi_seed_to_bo4e_mappings, lovric_expected",
        [
            pytest.param(
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator="$['Transaktionsgrund']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['EdiFoo']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                    ],
                ),
                {
                    "$['Transaktionsgrund']": "$['transaktionsdaten']['transaktionsgrund']",
                    "$['EdiFoo']": "$['stammdaten'][0]['applicationFoo']",
                },
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator="$['transaktionsdaten']['transaktionsgrund']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['stammdaten'][0]['applicationFoo']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )
    async def test_lovric_replace_edi_seed_paths_with_bo4e_paths(
        self,
        lovric_actual: ListOfValidationResultInContext,
        edi_seed_to_bo4e_mappings: dict,
        lovric_expected: ListOfValidationResultInContext,
    ):
        lovric_actual.replace_edi_domain_paths_with_application_domain_paths(edi_seed_to_bo4e_mappings)
        assert lovric_actual == lovric_expected

    @pytest.mark.parametrize(
        "lovric_actual, lovric_expected",
        [
            pytest.param(
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator="SG11",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="SG11 - Datum",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="SG11 - Datum - Auszug",
                            validation_result=DataElementValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                                format_validation_fulfilled=True,
                                data_element_data_type=DataElementDataType.DATETIME,
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="SG10 - Datum",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="SG10 - Datum - Einzug",
                            validation_result=DataElementValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                                format_validation_fulfilled=True,
                                data_element_data_type=DataElementDataType.DATETIME,
                            ),
                        ),
                    ],
                ),
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator="SG11 - Datum - Auszug",
                            validation_result=DataElementValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                                format_validation_fulfilled=True,
                                data_element_data_type=DataElementDataType.DATETIME,
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="SG10 - Datum - Einzug",
                            validation_result=DataElementValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_EMPTY,
                                format_validation_fulfilled=True,
                                data_element_data_type=DataElementDataType.DATETIME,
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )
    async def test_lovric_filter_for_data_element_validation_results(
        self, lovric_actual: ListOfValidationResultInContext, lovric_expected: ListOfValidationResultInContext
    ):
        lovric_actual.filter_for_data_element_validation_results()
        assert lovric_actual == lovric_expected

    @pytest.mark.parametrize(
        "lovric_actual, lovric_expected",
        [
            pytest.param(
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator="$['Transaktionsgrund']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['transaktionsdaten']['transaktionsgrund']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['stammdaten'][0]['applicationFoo']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['EdiFoo']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                    ],
                ),
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator="$['transaktionsdaten']['transaktionsgrund']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['stammdaten'][0]['applicationFoo']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )
    async def test_lovric_filter_for_boneycomb_path_results(
        self, lovric_actual: ListOfValidationResultInContext, lovric_expected: ListOfValidationResultInContext
    ):
        lovric_actual.filter_for_boneycomb_path_results()
        assert lovric_actual == lovric_expected

    @pytest.mark.parametrize(
        "lovric_actual, lovric_expected",
        [
            pytest.param(
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator='$["stammdaten"]["MARKTTEILNEHMER"]["ANSPRECHPARTNER"]["absender"]["ansprechpartner"]["nachname"]',
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['transaktionsdaten']['transaktionsgrund']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['stammdaten'][0]['applicationFoo']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator='$["stammdaten"]["MARKTTEILNEHMER"]["empfaenger"]["rollencodenummer"]',
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                    ],
                ),
                ListOfValidationResultInContext(
                    validation_results=[
                        ValidationResultInContext(
                            discriminator="$['transaktionsdaten']['transaktionsgrund']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                            ),
                        ),
                        ValidationResultInContext(
                            discriminator="$['stammdaten'][0]['applicationFoo']",
                            validation_result=SegmentLevelValidationResult(
                                requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                            ),
                        ),
                    ],
                ),
            ),
        ],
    )
    async def test_lovric_remove_absender_and_empfaenger_path_results(
        self, lovric_actual: ListOfValidationResultInContext, lovric_expected: ListOfValidationResultInContext
    ):
        lovric_actual.remove_absender_and_empfaenger_path_results()  # type:ignore[attr-defined]
        assert lovric_actual == lovric_expected



"""
Tests that the parsed trees are JSON serializable
"""
import json
import uuid
from typing import List, TypeVar

import pytest  # type:ignore[import]
from efoli import EdifactFormat
from marshmallow import Schema, ValidationError
from maus.models.edifact_components import DataElementDataType

from ahbicht.content_evaluation.categorized_key_extract import CategorizedKeyExtract, CategorizedKeyExtractSchema
from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
from ahbicht.evaluation_results import (
    AhbExpressionEvaluationResult,
    AhbExpressionEvaluationResultSchema,
    FormatConstraintEvaluationResult,
    RequirementConstraintEvaluationResult,
)
from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    EvaluatedFormatConstraint,
    EvaluatedFormatConstraintSchema,
)
from ahbicht.expressions.enums import ModalMark
from ahbicht.mapping_results import (
    ConditionKeyConditionTextMapping,
    ConditionKeyConditionTextMappingSchema,
    PackageKeyConditionExpressionMapping,
    PackageKeyConditionExpressionMappingSchema,
)
from ahbicht.validation.validation_results import (
    DataElementValidationResult,
    DataElementValidationResultSchema,
    ListOfValidationResultInContext,
    ListOfValidationResultInContextSchema,
    SegmentLevelValidationResult,
    SegmentLevelValidationResultSchema,
    ValidationResultInContext,
    ValidationResultInContextSchema,
)
from ahbicht.validation.validation_values import RequirementValidationValue

T = TypeVar("T")


def _test_serialization_roundtrip(serializable_object: T, schema: Schema, expected_json_dict: dict) -> T:
    """
    Serializes the serializable_object using the provided schema,
    asserts, that the result is equal to the expected_json_dict
    then deserializes it again and asserts on equality with the original serializable_object
    :returns the deserialized_object
    """
    json_string = schema.dumps(serializable_object)
    assert json_string is not None
    actual_json_dict = json.loads(json_string)
    assert actual_json_dict == expected_json_dict
    deserialized_object = schema.loads(json_data=json_string)
    assert isinstance(deserialized_object, type(serializable_object))
    assert deserialized_object == serializable_object
    return deserialized_object


class TestJsonSerialization:
    """
    This class tests the serialization of all objects except for trees.
    """

    @pytest.mark.parametrize(
        "evaluated_format_constraint, expected_json_dict",
        [
            pytest.param(
                EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                {"format_constraint_fulfilled": True, "error_message": None},
            ),
            pytest.param(
                EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="something is wrong"),
                {"format_constraint_fulfilled": False, "error_message": "something is wrong"},
            ),
        ],
    )
    def test_evaluated_format_constraint_serialization(
        self, evaluated_format_constraint: EvaluatedFormatConstraint, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            evaluated_format_constraint, EvaluatedFormatConstraintSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "invalid_content_evaluation_result_dict",
        [
            pytest.param({}, id="empty dict"),
            pytest.param({"format_constraints": {}, "hints": {}}, id="missing requirement constraints"),
            pytest.param({"requirement_constraints": {}, "hints": {}}, id="missing format_constraints"),
        ],
    )
    def test_validation_errors_on_content_evaluation_result_deserialization(
        self, invalid_content_evaluation_result_dict: dict
    ):
        schema = ContentEvaluationResultSchema()
        with pytest.raises(ValidationError):
            schema.load(invalid_content_evaluation_result_dict)

    @pytest.mark.parametrize(
        "content_evaluation_result, expected_json_dict",
        [
            pytest.param(
                ContentEvaluationResult(
                    hints={"501": "foo", "502": "bar", "503": None},
                    format_constraints={
                        "901": EvaluatedFormatConstraint(
                            format_constraint_fulfilled=False, error_message="something is wrong"
                        ),
                        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                    },
                    requirement_constraints={
                        "1": ConditionFulfilledValue.NEUTRAL,
                        "2": ConditionFulfilledValue.FULFILLED,
                        "3": ConditionFulfilledValue.UNFULFILLED,
                        "4": ConditionFulfilledValue.UNKNOWN,
                    },
                    packages={"123P": "[17] U [18]"},
                    id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
                ),
                {
                    "hints": {"501": "foo", "502": "bar", "503": None},
                    "format_constraints": {
                        "902": {"format_constraint_fulfilled": True, "error_message": None},
                        "901": {"format_constraint_fulfilled": False, "error_message": "something is wrong"},
                    },
                    "requirement_constraints": {"1": "NEUTRAL", "2": "FULFILLED", "3": "UNFULFILLED", "4": "UNKNOWN"},
                    "packages": {"123P": "[17] U [18]"},
                    "id": "d106f335-f663-4d14-9636-4f43a883ad26",
                },
            ),
        ],
    )
    def test_content_evaluation_result_serialization(
        self, content_evaluation_result: ContentEvaluationResult, expected_json_dict: dict
    ):
        for rc_evaluation_result in content_evaluation_result.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)
        deserialized_object = _test_serialization_roundtrip(
            content_evaluation_result, ContentEvaluationResultSchema(), expected_json_dict
        )
        for rc_evaluation_result in deserialized_object.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)
        for rc_evaluation_result in content_evaluation_result.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)

    def test_content_evaluation_result_without_packages_may_can_deserialized(self):
        json_dict = {
            "hints": {"501": "foo", "502": "bar", "503": None},
            "format_constraints": {
                "902": {"format_constraint_fulfilled": True, "error_message": None},
                "901": {"format_constraint_fulfilled": False, "error_message": "something is wrong"},
            },
            "requirement_constraints": {"1": "NEUTRAL", "2": "FULFILLED", "3": "UNFULFILLED", "4": "UNKNOWN"},
            "packages": None,
            "id": "d106f335-f663-4d14-9636-4f43a883ad26",
        }
        cer = ContentEvaluationResultSchema().load(json_dict)
        assert cer is not None
        assert cer.packages is None

    @pytest.mark.parametrize(
        "ahb_expression_evaluation_result, expected_json_dict",
        [
            pytest.param(
                AhbExpressionEvaluationResult(
                    requirement_indicator=ModalMark.MUSS,
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        error_message="hello", format_constraints_fulfilled=False
                    ),
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        hints="foo bar",
                        requirement_constraints_fulfilled=True,
                        requirement_is_conditional=True,
                        format_constraints_expression="[asd]",
                    ),
                ),
                {
                    "requirement_indicator": "MUSS",
                    "format_constraint_evaluation_result": {
                        "format_constraints_fulfilled": False,
                        "error_message": "hello",
                    },
                    "requirement_constraint_evaluation_result": {
                        "hints": "foo bar",
                        "requirement_is_conditional": True,
                        "requirement_constraints_fulfilled": True,
                        "format_constraints_expression": "[asd]",
                    },
                },
            ),
            pytest.param(
                AhbExpressionEvaluationResult(
                    requirement_indicator=ModalMark.MUSS,
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        error_message="hello", format_constraints_fulfilled=False
                    ),
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        hints="foo bar",
                        requirement_constraints_fulfilled=True,
                        requirement_is_conditional=True,
                        format_constraints_expression="[asd]",
                    ),
                ),
                {
                    "format_constraint_evaluation_result": {
                        "error_message": "hello",
                        "format_constraints_fulfilled": False,
                    },
                    "requirement_constraint_evaluation_result": {
                        "format_constraints_expression": "[asd]",
                        "hints": "foo bar",
                        "requirement_constraints_fulfilled": True,
                        "requirement_is_conditional": True,
                    },
                    "requirement_indicator": "MUSS",
                },
            ),
            pytest.param(
                AhbExpressionEvaluationResult(
                    requirement_indicator=ModalMark.MUSS,
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        format_constraints_fulfilled=False
                    ),
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        requirement_constraints_fulfilled=True,
                        requirement_is_conditional=True,
                    ),
                ),
                {
                    "format_constraint_evaluation_result": {
                        "error_message": None,
                        "format_constraints_fulfilled": False,
                    },
                    "requirement_constraint_evaluation_result": {
                        "format_constraints_expression": None,
                        "hints": None,
                        "requirement_constraints_fulfilled": True,
                        "requirement_is_conditional": True,
                    },
                    "requirement_indicator": "MUSS",
                },
                id="Minimal example",
            ),
        ],
    )
    def test_ahb_expression_evaluation_result_serialization(
        self, ahb_expression_evaluation_result: AhbExpressionEvaluationResult, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            ahb_expression_evaluation_result, AhbExpressionEvaluationResultSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "validation_result, expected_json_dict",
        [
            pytest.param(
                SegmentLevelValidationResult(
                    requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                ),
                {
                    "requirement_validation": "IS_FORBIDDEN_AND_EMPTY",
                    "hints": "foo",
                },
            ),
            pytest.param(
                SegmentLevelValidationResult(requirement_validation=RequirementValidationValue.IS_FORBIDDEN),
                {
                    "requirement_validation": "IS_FORBIDDEN",
                    "hints": None,
                },
            ),
        ],
    )
    def test_segment_level_validation_result_serialization(
        self, validation_result: SegmentLevelValidationResult, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(validation_result, SegmentLevelValidationResultSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "validation_result, expected_json_dict",
        [
            pytest.param(
                DataElementValidationResult(
                    requirement_validation=RequirementValidationValue.IS_FORBIDDEN,
                    format_validation_fulfilled=True,
                ),
                {
                    "data_element_data_type": None,
                    "requirement_validation": "IS_FORBIDDEN",
                    "hints": None,
                    "format_validation_fulfilled": True,
                    "format_error_message": None,
                    "possible_values": None,
                },
            ),
            pytest.param(
                DataElementValidationResult(
                    requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_FILLED,
                    hints="foo",
                    format_validation_fulfilled=False,
                    format_error_message="bar",
                ),
                {
                    "data_element_data_type": None,
                    "requirement_validation": "IS_REQUIRED_AND_FILLED",
                    "hints": "foo",
                    "format_validation_fulfilled": False,
                    "format_error_message": "bar",
                    "possible_values": None,
                },
            ),
            pytest.param(
                DataElementValidationResult(
                    requirement_validation=RequirementValidationValue.IS_REQUIRED,
                    hints="foo",
                    format_validation_fulfilled=True,
                    format_error_message=None,
                    possible_values={"A1": "Ich bin A1", "A2": "Ich bin A2", "Z3": "Ich bin Z3"},
                ),
                {
                    "data_element_data_type": None,
                    "requirement_validation": "IS_REQUIRED",
                    "hints": "foo",
                    "format_validation_fulfilled": True,
                    "format_error_message": None,
                    "possible_values": {"A1": "Ich bin A1", "A2": "Ich bin A2", "Z3": "Ich bin Z3"},
                },
            ),
            pytest.param(
                DataElementValidationResult(
                    requirement_validation=RequirementValidationValue.IS_REQUIRED,
                    hints="foo",
                    format_validation_fulfilled=True,
                    format_error_message=None,
                    possible_values={"A1": "Ich bin A1", "A2": "Ich bin A2", "Z3": "Ich bin Z3"},
                    data_element_data_type=DataElementDataType.VALUE_POOL,
                ),
                {
                    "requirement_validation": "IS_REQUIRED",
                    "hints": "foo",
                    "format_validation_fulfilled": True,
                    "format_error_message": None,
                    "possible_values": {"A1": "Ich bin A1", "A2": "Ich bin A2", "Z3": "Ich bin Z3"},
                    "data_element_data_type": "VALUE_POOL",
                },
            ),
        ],
    )
    def test_data_element_validation_result_serialization(
        self, validation_result: DataElementValidationResult, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(validation_result, DataElementValidationResultSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "validation_result_in_context, expected_json_dict",
        [
            pytest.param(
                ValidationResultInContext(
                    discriminator="foo",
                    validation_result=SegmentLevelValidationResult(
                        requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                    ),
                ),
                {
                    "discriminator": "foo",
                    "validation_result": {
                        "requirement_validation": "IS_FORBIDDEN_AND_EMPTY",
                        "hints": "foo",
                    },
                },
            ),
            pytest.param(
                ValidationResultInContext(
                    discriminator="foo_dataelement",
                    validation_result=DataElementValidationResult(
                        requirement_validation=RequirementValidationValue.IS_REQUIRED_AND_FILLED,
                        hints="foo",
                        format_validation_fulfilled=False,
                        format_error_message="bar",
                    ),
                ),
                {
                    "discriminator": "foo_dataelement",
                    "validation_result": {
                        "data_element_data_type": None,
                        "requirement_validation": "IS_REQUIRED_AND_FILLED",
                        "hints": "foo",
                        "format_validation_fulfilled": False,
                        "format_error_message": "bar",
                        "possible_values": None,
                    },
                },
            ),
        ],
    )
    def test_validation_result_in_context_serialization(
        self, validation_result_in_context: ValidationResultInContext, expected_json_dict: dict
    ):
        json_string = ValidationResultInContextSchema().dumps(validation_result_in_context)
        assert json_string is not None
        actual_json_dict = json.loads(json_string)
        assert actual_json_dict == expected_json_dict

    @pytest.mark.parametrize(
        "list_of_validation_result_in_context, expected_json_dict",
        [
            pytest.param(
                [
                    ValidationResultInContext(
                        discriminator="foo",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_FORBIDDEN_AND_EMPTY, hints="foo"
                        ),
                    ),
                    ValidationResultInContext(
                        discriminator="bar",
                        validation_result=SegmentLevelValidationResult(
                            requirement_validation=RequirementValidationValue.IS_REQUIRED, hints="bar"
                        ),
                    ),
                ],
                [
                    {
                        "discriminator": "foo",
                        "validation_result": {
                            "requirement_validation": "IS_FORBIDDEN_AND_EMPTY",
                            "hints": "foo",
                        },
                    },
                    {
                        "discriminator": "bar",
                        "validation_result": {
                            "requirement_validation": "IS_REQUIRED",
                            "hints": "bar",
                        },
                    },
                ],
            ),
        ],
    )
    def test_list_of_validation_result_in_context_serialization(
        self,
        list_of_validation_result_in_context: List[ValidationResultInContext],
        expected_json_dict: dict,
    ):
        json_string = ValidationResultInContextSchema().dumps(list_of_validation_result_in_context, many=True)
        assert json_string is not None
        actual_json_dict = json.loads(json_string)
        assert actual_json_dict == expected_json_dict

        list_of_validation_result_in_context_as_class = ListOfValidationResultInContext(
            validation_results=list_of_validation_result_in_context
        )
        json_string = ValidationResultInContextSchema().dumps(
            list_of_validation_result_in_context_as_class.validation_results, many=True
        )
        assert json_string is not None
        actual_json_dict = json.loads(json_string)
        assert actual_json_dict == expected_json_dict

    @pytest.mark.parametrize(
        "condition_key_condition_text_mapping, expected_json_dict",
        [
            pytest.param(
                ConditionKeyConditionTextMapping(
                    edifact_format=EdifactFormat.UTILMD,
                    condition_key="123",
                    condition_text="Blablabla",
                ),
                {"edifact_format": "UTILMD", "condition_key": "123", "condition_text": "Blablabla"},
            ),
        ],
    )
    def test_condition_key_condition_text_mapping_serialization(
        self, condition_key_condition_text_mapping: ConditionKeyConditionTextMapping, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            condition_key_condition_text_mapping, ConditionKeyConditionTextMappingSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "package_key_condition_expression_mapping, expected_json_dict",
        [
            pytest.param(
                PackageKeyConditionExpressionMapping(
                    edifact_format=EdifactFormat.UTILMD,
                    package_key="123P",
                    package_expression="[1] U [2] O [3]",
                ),
                {"edifact_format": "UTILMD", "package_key": "123P", "package_expression": "[1] U [2] O [3]"},
            ),
        ],
    )
    def test_package_key_condition_expression_mapping_serialization(
        self, package_key_condition_expression_mapping: PackageKeyConditionExpressionMapping, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            package_key_condition_expression_mapping, PackageKeyConditionExpressionMappingSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "categorized_key_extract, expected_json_dict",
        [
            pytest.param(
                CategorizedKeyExtract(
                    hint_keys=["501", "502", "503"],
                    format_constraint_keys=["901", "902"],
                    requirement_constraint_keys=["1", "2", "3", "4"],
                    package_keys=["17P"],
                    time_condition_keys=["UB1", "UB2"],
                ),
                {
                    "hint_keys": ["501", "502", "503"],
                    "format_constraint_keys": ["901", "902"],
                    "requirement_constraint_keys": ["1", "2", "3", "4"],
                    "package_keys": ["17P"],
                    "time_condition_keys": ["UB1", "UB2"],
                },
            ),
        ],
    )
    def test_categorized_key_extract_serialization(
        self, categorized_key_extract: CategorizedKeyExtract, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(categorized_key_extract, CategorizedKeyExtractSchema(), expected_json_dict)
